/***************************************************************************\
*****************************************************************************
**                                                                         **
** source file:    server.c                                                **
** last revision:  24/02/2004                                              **
**                                                                         **
*****************************************************************************
*****************************************************************************
**                                                                         **
** This program is free software; you can redistribute it and/or modify it **
** under the terms of the GNU General Public License as published by the   **
** Free Software Foundation; either version 2, or (at your option) any     **
** later version.                                                          **
**                                                                         **
** This program is distributed in the hope that it will be useful,         **
** but WITHOUT ANY WARRANTY; without even the implied warranty of          **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           **
** GNU General Public License for more details.                            **
**                                                                         **
** You should have received a copy of the GNU General Public License       **
** along with this program; if not, write to the Free Software             **
** Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.                     **
**                                                                         **
*****************************************************************************
\***************************************************************************/

//--------------------------------------------------------------------------------------------
//||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//--------------------------------------------------------------------------------------------

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include "common.h"
#include "file_io.h"
#include "error.h"
#include "socket.h"
#include "utils.h"
#include "memory.h"
#include "server.h"

//--------------------------------------------------------------------------------------------
//||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//--------------------------------------------------------------------------------------------

int start_server(int port, long int *limit, MEM_MANAGER *mem_manager)
{
  int my_socket;                    // The Socket
  struct sockaddr_in my_addr;       // The Server SockAddr
  struct sockaddr_in client;        // The Client SockAddr
  struct file_t *rcv_file;          // The Files We'll Receive
  int nbr_files;                    // The Number Of Files We're Going To Receive
  int i, j, k;                      // Storage For Loops And Temp Ints
  int optval = 1;                   // Storage For Some Options
  int dial_sock;                    // The Socket For The Dialog With The Client
  int bytes_we_have;                // The Number Of Bytes We Have During A Transfer
  int bytes_received;               // The Number Of Bytes We Received During A Transfer
  size_t file_listing_size;         // The Size Of The File Listing
  char have_limit = 0;              // Do We Have A Limit?
  char *dial_string;                // This Will Contain The Dialog Strings
  char *tmp_string;                 // This Will Be A Storage For Temp Strings
  char *file_buffer;                // This Will Be A Storage For The File Buffer
  char *ptr_buffer;                 // This Is A Pointer To A Buffer
  struct timeval time_current;      // The Current Time
  struct timeval time_last_update;  // The Time For The Next Update Of The Kbps Counter
  float speed;                      // The Transfer's Speed
  float new_speed;                  // The Next Transfer's Speed We'll Display

  /** We Allocate Some Memory **/
  if ((mem_alloc(mem_manager, (void **) &dial_string, DIAL_BUFFER_SIZE)) != 0)
    return -1;

  if ((mem_alloc(mem_manager, (void **) &tmp_string, DIAL_BUFFER_SIZE)) != 0)
    return -1;

  /** We Create A Socket **/
  if ((my_socket = create_socket()) < 0)
      return -1;

  /** We Set Some Options **/
  if (setsockopt(my_socket, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(int)) != 0)
    {
      error("setsockopt", NULL, errno);
      return -1;
    }

  memset(&client, 0, sizeof(struct sockaddr_in));
  memset(&my_addr, 0, sizeof(struct sockaddr_in));
  my_addr.sin_family = AF_INET;
  my_addr.sin_port = htons(port);
  my_addr.sin_addr.s_addr = htonl(INADDR_ANY);

  if (*limit > 0)
    have_limit = 1;

  if (bind(my_socket, (struct sockaddr *) &my_addr, sizeof(struct sockaddr_in)) != 0)
    {
      error("bind", NULL, errno);
      return -1;
    }

  /** We Prepare The Socket To Listen **/
  if (listen(my_socket, 1) != 0)
    {
      error("listen", NULL, errno);
      return -1;
    }

  /** We Wait For A Connection **/
  fprintf(stderr, "Waiting for a connection...\n");
  optval = sizeof(struct sockaddr_in);

  if ((dial_sock = accept(my_socket, (struct sockaddr *) &client, &optval)) < 0)
    {
      error("accept", NULL, errno);
      return -1;
    }

  /** Our Socket Must Not Block **/
  if ((fcntl(dial_sock, F_SETFL, O_NONBLOCK)) < 0)
    {
      error("fcntl", NULL, errno);
      return stop_server(dial_sock);
    }

  /** We Don't Want To Listen Anymore **/
  if ((close_socket(my_socket)) < 0)
    return stop_server(dial_sock);

  /** We Print Some Infos About The Client **/
  fprintf(stderr, "Client %s connected.\n", (inet_ntoa(client.sin_addr)));

#ifdef WITH_UNAME
  /** We Send The Welcome String To The Client **/
  if (get_sys_infos(tmp_string) == 0)
    sprintf(dial_string, "#Welcome on this %s server version %s\n" \
                         "#Server is running %s\n" \
                         "#Now ready to receive the files...\n", PROGRAM_NAME, PROGRAM_VERSION, tmp_string);
  else
    sprintf(dial_string, "#Welcome on this %s server version %s\n" \
                         "#Server is now ready to receive the files...\n", PROGRAM_NAME, PROGRAM_VERSION);
#else
  sprintf(dial_string, "#Welcome on this %s server version %s\n" \
                       "#Server is now ready to receive the files...\n", PROGRAM_NAME, PROGRAM_VERSION);
#endif
  
  if ((write_to_socket(dial_sock, dial_string, 1+strlen(dial_string))) < 0)
    return stop_server(dial_sock);

  /** We Wait For An Answer (Size Of File Listing) **/
  if ((wait_for_input(dial_sock, TIME_OUT)) != 0)
    return stop_server(dial_sock);

  if ((read_from_socket(dial_sock, dial_string, mem_sizeof(mem_manager, dial_string))) < 0)
    return stop_server(dial_sock);

  /** We Reallocate Some Memory If We Have To **/
  file_listing_size = atoi(dial_string);

  if (file_listing_size == 0)
    {
      error("start_server", "invalid transmission", 0);
      return stop_server(dial_sock);
    }

  if (file_listing_size > LISTING_MAX_SIZE)
    {
      error("start_server", "file listing is too long (limitation removal might be expected in 0.6.0)", 0);
      write_to_socket(dial_sock, STR_TOO_LONG, 1+strlen(STR_TOO_LONG));
      shutdown(dial_sock, SHUT_RDWR);
      close(dial_sock);
      return -1;
    }

  if ((mem_sizeof(mem_manager, dial_string)) < file_listing_size)
    {
      if ((mem_realloc(mem_manager, (void **) &dial_string, file_listing_size)) != 0)
        return stop_server(dial_sock);
    }

  if ((mem_sizeof(mem_manager, tmp_string)) < file_listing_size)
    {
      if ((mem_realloc(mem_manager, (void **) &tmp_string, file_listing_size)) != 0)
        return stop_server(dial_sock);
    }

  memset(dial_string, 0, mem_sizeof(mem_manager, dial_string));
  memset(tmp_string, 0, mem_sizeof(mem_manager, tmp_string));

  /** We Read The File Listing **/
  bytes_we_have = 0;
  bytes_received = 0;
  ptr_buffer = dial_string;

  while (bytes_we_have < file_listing_size)
    {
      if ((write_to_socket(dial_sock, STR_CONFIRM, 1+strlen(STR_CONFIRM))) < 0)
        return stop_server(dial_sock);

      if ((wait_for_input(dial_sock, TIME_OUT)) != 0)
        return stop_server(dial_sock);

      if ((bytes_received = read_from_socket(dial_sock, ptr_buffer, (file_listing_size - bytes_we_have))) < 0)
        return stop_server(dial_sock);

      ptr_buffer += bytes_received;      // We Update The File Buffer Pointer
      bytes_we_have += bytes_received;   // We Update The Number Of Bytes Of Datas We Have
    }

  if ((1 + strlen(dial_string)) != file_listing_size)
    {
      error("start_server", "the file listing I have is invalid", 0);
      return stop_server(dial_sock);
    }

  /** We Now Have To Analyse The File Listing **/
  // Number Of Files
  j = 0;
  k = 0;
  while (dial_string[k] != '\n')  ++k; // We Go At The End Of The Number Of Files Marked By \n
  memcpy(tmp_string, &dial_string[j], k-j);
  tmp_string[k-j] = 0;
  j += (k-j+1);
  nbr_files = atoi(tmp_string);

  // Memory Allocation For Files Struct
  if ((mem_alloc(mem_manager, (void **) &rcv_file, (nbr_files * sizeof(struct file_t)))) != 0)
    return stop_server(dial_sock);

  for (i=0; i<nbr_files; ++i)
    {
      // File Name
      k = j;
      while (dial_string[k] != '\n')  ++k; // We Go At The End Of The File Name Marked By \n
      memcpy(tmp_string, &dial_string[j], k-j);
      tmp_string[k-j] = 0;
      j += (k-j+1);
      if ((mem_alloc(mem_manager, (void **) &rcv_file[i].name, k)) != 0)  return stop_server(dial_sock);
      strcpy(rcv_file[i].name, tmp_string);

      // File Size
      k = j;
      while (dial_string[k] != '\n')  ++k; // We Go At The End Of The Size Marked By \n
      memcpy(tmp_string, &dial_string[j], k-j);
      tmp_string[k-j] = 0;
      j += (k-j+1);
      rcv_file[i].size = (size_t) atoi(tmp_string);

      // File Checksum
      k = j;
      while (dial_string[k] != '\n')  ++k; // We Go At The End Of The Checksum Marked By \n
      memcpy(tmp_string, &dial_string[j], k-j);
      tmp_string[k-j] = 0;
      j += (k-j+1);
      rcv_file[i].checksum =  atol(tmp_string);
    }

  /** We Reallocate The DialString **/
  if ((mem_realloc(mem_manager, (void **) &dial_string, DIAL_BUFFER_SIZE)) != 0)
    return stop_server(dial_sock);

  /** We Allocate Some Memory For The File Buffer **/
  if ((mem_alloc(mem_manager,  (void **) &file_buffer, FILE_BUFFER_SIZE)) != 0)
	return stop_server(dial_sock);

  memset(dial_string, 0, mem_sizeof(mem_manager, dial_string));
  memset(tmp_string, 0, mem_sizeof(mem_manager, tmp_string));

  /** We Receive The Files!!! **/
  for (i=0; i<nbr_files; ++i)
    {
      int out_file;
      int bytes_written = 0;
      long checksum = 0;

      bytes_we_have = 0;
      bytes_received = 0;

      if (have_limit)
        {
          *limit -= rcv_file[i].size;

          if (*limit < 0)
            {
              write_to_socket(dial_sock, STR_LIMIT, 1+strlen(STR_LIMIT));
              error("start_server", "bytes limit is about to be exceeded on server side, bailing out", 0);
              return stop_server(dial_sock);
            }
        }

      if ((out_file = open_file(rcv_file[i].name, (O_WRONLY | O_CREAT | O_TRUNC))) < 0)
        return stop_server(dial_sock);

      ptr_buffer = file_buffer;
      gettimeofday(&time_current, NULL);
      time_last_update = time_current;
      speed = -1.0f;
      new_speed = 0.0f;
      fprintf(stderr, "\nReceiving %s 0%%", rcv_file[i].name);

      while (bytes_written < rcv_file[i].size)
        {
          if ((write_to_socket(dial_sock, STR_CONFIRM, 1+strlen(STR_CONFIRM))) < 0)
            return stop_server(dial_sock);

          if ((wait_for_input(dial_sock, TIME_OUT)) != 0)
            return stop_server(dial_sock);

          if ((rcv_file[i].size - bytes_we_have) > DIAL_BUFFER_SIZE)
            {
              if ((bytes_received = read_from_socket(dial_sock, dial_string, DIAL_BUFFER_SIZE)) < 0)
                return stop_server(dial_sock);
            }
          else
            {
              if ((bytes_received = read_from_socket(dial_sock, dial_string, (rcv_file[i].size - bytes_we_have))) < 0)
                return stop_server(dial_sock);
            }

          checksum += get_buf_checksum(dial_string, bytes_received);     // We Update The Checksum
          memcpy(ptr_buffer, dial_string, bytes_received);               // We Copy Datas To The File Buffer
          ptr_buffer += bytes_received;                                  // We Update The File Buffer Pointer
          bytes_we_have = (bytes_written + (ptr_buffer - file_buffer));  // We Update The Number Of Bytes Of Datas We Have (written or buffered)
          new_speed += (float) bytes_received;

          if (((ptr_buffer - file_buffer) > (FILE_BUFFER_SIZE - DIAL_BUFFER_SIZE)) || (bytes_we_have >= rcv_file[i].size))
            {
              if ((write_to_file(out_file, file_buffer, (ptr_buffer - file_buffer))) < 0)
                return stop_server(dial_sock);

              bytes_written += (ptr_buffer - file_buffer);
              ptr_buffer = file_buffer;
            }

          gettimeofday(&time_current, NULL);

          if (time_current.tv_sec > time_last_update.tv_sec)
            {
              float elapsed_time = (time_current.tv_sec - time_last_update.tv_sec) + ((time_current.tv_usec - time_last_update.tv_usec) / 1000000.0f);

              if (elapsed_time >= 2.0f)
                {
                  new_speed = ((new_speed  / elapsed_time) / 1000.0f);
                  speed = ((speed + new_speed) / 2.0f);
                  new_speed = 0.0f;
                  gettimeofday(&time_last_update, NULL);
                }
            }

          if (speed > 0.0f)
            fprintf(stderr, "\r\t                                                             \rReceiving %s %d%% (%.1f kbps)", rcv_file[i].name,(int) (((float) bytes_we_have / (float) rcv_file[i].size) * 100.0f), speed);
          else
            fprintf(stderr, "\r\t                                                             \rReceiving %s %d%% (-.- kbps)", rcv_file[i].name,(int) (((float) bytes_we_have / (float) rcv_file[i].size) * 100.0f));
        }

      if ((close(out_file)) < 0)
        return stop_server(dial_sock);

      if (checksum != rcv_file[i].checksum)
        {
          error("start_server", "invalid file checksum, bailing out", 0);
          return stop_server(dial_sock);
        }

      fprintf(stderr, "\r\t                                                             \rReceiving %s 100%%", rcv_file[i].name);
    }

  fprintf(stderr, "\n\nAll the files have been received successfully.\n\n");

  /** We Send The Bye Bye String To The Client **/  
  if ((write_to_socket(dial_sock, STR_QUIT, 1+strlen(STR_QUIT))) < 0)
      return -1;

  sleep(1);

  /** We Close The Dialog Socket **/
  if ((close_socket(dial_sock)) != 0)
    return -1;

  return 0;
}

//--------------------------------------------------------------------------------------------
//||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//--------------------------------------------------------------------------------------------

int stop_server(int socket)
{
  write_to_socket(socket, STR_STOP, 1+strlen(STR_STOP));
  shutdown(socket, SHUT_RDWR);
  close(socket);
  return -1;
}

//--------------------------------------------------------------------------------------------
//||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//--------------------------------------------------------------------------------------------
