/***************************************************************************\
*****************************************************************************
**                                                                         **
** source file:    socket.c                                                **
** last revision:  21/02/2004                                              **
**                                                                         **
*****************************************************************************
*****************************************************************************
**                                                                         **
** This program is free software; you can redistribute it and/or modify it **
** under the terms of the GNU General Public License as published by the   **
** Free Software Foundation; either version 2, or (at your option) any     **
** later version.                                                          **
**                                                                         **
** This program is distributed in the hope that it will be useful,         **
** but WITHOUT ANY WARRANTY; without even the implied warranty of          **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           **
** GNU General Public License for more details.                            **
**                                                                         **
** You should have received a copy of the GNU General Public License       **
** along with this program; if not, write to the Free Software             **
** Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.                     **
**                                                                         **
*****************************************************************************
\***************************************************************************/

//---------------------------------------------------------------------------------------
//|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//---------------------------------------------------------------------------------------

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/time.h>
#include "common.h"
#include "memory.h"
#include "error.h"
#include "socket.h"

//---------------------------------------------------------------------------------------
//|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//---------------------------------------------------------------------------------------

int create_socket(void)
{
  int new_socket;

  new_socket = socket(PF_INET, SOCK_STREAM, 0);
  
  if (new_socket == -1)
    error("socket", NULL, errno);

  return new_socket;
}


//---------------------------------------------------------------------------------------
//|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//---------------------------------------------------------------------------------------

int wait_for_input(const int socket, const int time)
{
  fd_set rfds;                 // Stream Reference For select()
  struct timeval tv;           // Time Value For The select() Function
  int retval;                  // Return Value Of Select

  memset(&tv, 0, sizeof(struct timeval));

  FD_ZERO(&rfds);
  FD_SET(socket, &rfds);

  tv.tv_sec = time;
  tv.tv_usec = 0;

  retval = select(socket+1, &rfds, NULL, NULL, &tv);

  if (retval == -1)
    {
      error("select", NULL, errno);
      return -1;
    }
  else if (retval == 0)
    {
      fprintf(stderr, "\n%s: error: Time Out!\n\n", PROGRAM_NAME);
      return -1;
    }

  return 0;
}


//---------------------------------------------------------------------------------------
//|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//---------------------------------------------------------------------------------------

int get_confirmation(const int socket, char *buffer, const size_t len)
{
  if (buffer == NULL)
    return -1;

  if ((wait_for_input(socket, TIME_OUT)) != 0)
    return -1;

  if ((read_from_socket(socket, buffer, len)) < 0)
    return -1;

  if ((strcmp(buffer, STR_CONFIRM) != 0))
    {
      if (buffer[0] == '+')
        {
          fprintf(stderr, "\n\n%s\n", buffer);
          return -1;
        }
      else
        {
          fprintf(stderr, "%s: get_confirmation error: No confirmation packet received.\n\n", PROGRAM_NAME);
          return -1;
        }
    }

  return 0;
}


//---------------------------------------------------------------------------------------
//|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//---------------------------------------------------------------------------------------

size_t read_from_socket(const int socket, void *buffer, const size_t size)
{
  int len_read = 0;
  void *ptr = buffer;

  while ((ptr - buffer) < size)
    {
      len_read = read(socket, ptr, size - (ptr - buffer));

      if (len_read < 0)
        {
          if (errno == EAGAIN) // We Skip This Error Because We Are In Non-Block Mode
            {
              return (ptr - buffer);
            }
          else
            {	   
              error("read_from_socket", NULL, errno);
              return -1;
            }
        }
      else if (len_read == 0)
        {
          return (ptr - buffer);
        }

      ptr += len_read;
    }

  return (ptr - buffer);
}


//---------------------------------------------------------------------------------------
//|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//---------------------------------------------------------------------------------------

size_t write_to_socket(const int socket, void *buffer, const size_t size)
{
  int len_written = 0;
  void *ptr = buffer;
  
  while ((ptr - buffer) < size)
    {
      len_written = write(socket, ptr, size - (ptr - buffer));

      if (len_written < 0)
        {
          if (errno == EAGAIN) // We Skip This Error Because We Are In Non-Block Mode
            {
              return (ptr - buffer);
            }
          else
            {	   
              error("write_to_socket", NULL, errno);
              return -1;
            }

          return -1;
        }
      else if (len_written == 0)
        {
          return (ptr - buffer);
        }

      ptr += len_written;
    }

  return (ptr - buffer);
}


//---------------------------------------------------------------------------------------
//|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//---------------------------------------------------------------------------------------

int close_socket(const int socket)
{
  shutdown(socket, SHUT_RDWR); // We Don't Care Here About Possible Errors As We're Calling That Just For Fun ;)
  
  if ((close(socket)) != 0)
    {
      error("close_socket", NULL, errno);
      return -1;
    }

  return 0;
}


//---------------------------------------------------------------------------------------
//|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//---------------------------------------------------------------------------------------
